from xclientai.utils.requests_utils import do_request
from xclientai.config import Config

# This function is not in config utils to avoid circular imports
def is_config_valid(api_key: str, workspace_id: str):        
    response = do_request(
        url=f"{Config.AUTH_BASE_URL_X_BACKEND}/v1/auth/verify",
        http_method="post",
        headers={
            "x-api-key": api_key,
            # Every user should have at least read permissions
            "x-original-method": "GET",
            "x-original-url": "https://dev.xcloud-api.stochastic.ai/executor/backend/v1/jobs/"
        },
        xcloud_auth=False,
        workspace_id=workspace_id,
        throw_error=False
    )
    
    try:
        response.raise_for_status()
        return True
    except:
        return False


from typing import List, Any, Dict, Optional, Union
from pydantic import BaseModel, validator, Field
from enum import Enum
from xclientai.utils.logging import configure_logger
from xclientai.config import Config
from pathlib import Path
from urllib.parse import urlparse
import json


logger = configure_logger(__name__)

class Metadata(BaseModel):
    error: Optional[str]
    ended_at: Optional[int]
    started_at: Optional[int]
    started_reason: Optional[str]
    cancelled_at: Optional[str]
    cancelled_reason: Optional[str]
    deleted_at: Optional[str]
    deleted_reason: Optional[str]
    
    class Config:
        validate_all = True

class MachineType(str, Enum):
    CPU_SMALL = "cpu_small"
    CPU_MEDIUM = "cpu_medium"
    CPU_LARGE = "cpu_large"
    CPU_XLARGE = "cpu_xlarge"
    GPU_T4_1 = "gpu_t4_1"
    GPU_T4_2 = "gpu_t4_2"
    GPU_T4_4 = "gpu_t4_4"
    GPU_L4_1 = "gpu_l4_1"
    GPU_L4_2 = "gpu_l4_2"
    GPU_L4_4 = "gpu_l4_4"
    GPU_L4_8 = "gpu_l4_8"
    GPU_V100_1 = "gpu_v100_1"
    GPU_V100_2 = "gpu_v100_2"
    GPU_V100_4 = "gpu_v100_4"
    GPU_V100_8 = "gpu_v100_8"
    GPU_A100_40_1 = "gpu_a100_40_1"
    GPU_A100_40_2 = "gpu_a100_40_2"
    GPU_A100_40_4 = "gpu_a100_40_4"
    GPU_A100_40_8 = "gpu_a100_40_8"
    GPU_A100_40_16 = "gpu_a100_40_16"
    GPU_A100_80_1 = "gpu_a100_80_1"
    GPU_A100_80_2 = "gpu_a100_80_2"
    GPU_A100_80_4 = "gpu_a100_80_4"
    GPU_A100_80_8 = "gpu_a100_80_8"   

class Status(str, Enum):
    SUCCESSFUL = "successful"
    FAILED = "failed"
    RUNNING = "running"
    NOT_STARTED = "not_started"
    CANCELLED = "cancelled"
    DEPLOYING = "deploying"
    
class Credentials(BaseModel):
    aws_access_key_id: Optional[str] = Field(repr=False)
    aws_secret_access_key: Optional[str] = Field(repr=False)
    aws_region: Optional[str] = Field(repr=False)
    application_credentials: Optional[Union[str, Dict]] = Field(repr=False)
    cloud: str
    
    @validator("aws_access_key_id")
    def validate_aws_access_key_id(cls, v):
        return v
    
    @validator("aws_secret_access_key")
    def validate_aws_secret_access_key(cls, v):
        return v
    
    @validator("aws_region")
    def validate_aws_region(cls, v):
        return v
    
    @validator("application_credentials")
    def validate_application_credentials(cls, v):
        if v is not None and isinstance(v, str):
            application_credentials_path = Path(v)
            assert application_credentials_path.is_file(), "application_credentials should be a path to a file"
            
            with open(str(application_credentials_path.resolve())) as f:
                application_credentials_content = json.loads(f.read())
        
            return application_credentials_content
        return v
    
    @validator("cloud")
    def validate_cloud(cls, v, values, **kwargs):
        valid_clouds = ["aws", "gcp", None]
        
        assert v in valid_clouds, f"Valid clouds are {valid_clouds}"
        
        if v == "aws":
            condition = "aws_access_key_id" in values and "aws_secret_access_key" in values and "aws_region" in values
            
            if not condition:
                logger.warning("AWS credentials have not been provided")
                
        if v == "gcp":
            condition = "application_credentials" in values
            
            if not condition:
                logger.warning("Google application_credentials have not been provided")
        
        return v
    
    class Config:
        validate_all = True 



import click
from typing import Optional

from xclientai.config import Config
from xclientai.utils.logging import configure_logger
from xclientai.utils.config_utils import overwrite_config
from xclientai.utils.validations import is_config_valid

logger = configure_logger(__name__)

@click.command(name="configure")
@click.option('-k', '--api_key', hide_input=True, prompt='[*] Your API key')
@click.option('-w', '--default_workspace', prompt='[*] Your default workspace')
def configure_command(api_key: Optional[str], default_workspace: Optional[str]):    
    xcloud_config = {
        "XCLOUD_API_KEY": api_key,
        "XCLOUD_DEFAULT_WORKSPACE": default_workspace
    }
        
    click.secho('[*] Validating your data...', fg='white', bold=True)
    is_valid = is_config_valid(api_key=api_key, workspace_id=default_workspace)
    if not is_valid:
        return click.secho('[!] Your configuration is not valid. Make sure to provide the correct API Key and default workspace ID. If the problem persists contact support.', fg='red', bold=True)
    
    overwrite_config(xcloud_config)
    click.secho('[+] Configuration added successfully', fg='green', bold=True)


from xclientai.utils.requests_utils import do_request
from xclientai.config import Config
from xclientai.dtypes.executor import ExecutionJob
from typing import List, Optional


class ExecutionJobsClient:
    
    @classmethod
    def get_jobs(cls, workspace_id: Optional[str] = None) -> List[ExecutionJob]:        
        response = do_request(
            url="{}/v1/jobs/".format(
                Config.EXECUTION_BASE_URL_X_BACKEND
            ),
            http_method="get",
            workspace_id=workspace_id
        )
        
        list_job_dicts = response.json()["data"]
        jobs = []
        
        for job_dict in list_job_dicts:
            job = ExecutionJob.parse_obj(job_dict)
            jobs.append(job)
            
        return jobs
    
    @classmethod
    def get_job_by_name(cls, job_name: str, workspace_id: Optional[str] = None) -> ExecutionJob:        
        response = do_request(
            url="{}/v1/jobs/{}".format(
                Config.EXECUTION_BASE_URL_X_BACKEND,
                job_name
            ),
            http_method="get",
            workspace_id=workspace_id
        )
        
        job_dict = response.json()["data"]
        job = ExecutionJob.parse_obj(job_dict)
        
        return job
    
    @classmethod
    def get_logs(cls, job_name: str, workspace_id: Optional[str] = None) -> List[ExecutionJob]:        
        response = do_request(
            url="{}/v1/jobs/{}/logs".format(
                Config.EXECUTION_BASE_URL_X_BACKEND,
                job_name
            ),
            http_method="get",
            workspace_id=workspace_id
        )
        
        logs = response.json()["data"]
        
        return logs
    
    @classmethod
    def cancel_job(cls, job_name: str, reason: str = "", workspace_id: Optional[str] = None) -> ExecutionJob:
        response = do_request(
            url="{}/v1/jobs/{}".format(
                Config.EXECUTION_BASE_URL_X_BACKEND,
                job_name
            ),
            params={
                "hard_delete": False,
                "reason": reason
            },
            http_method="delete",
            workspace_id=workspace_id
        )
        
        job_dict = response.json()["data"]
        job = ExecutionJob.parse_obj(job_dict)
        return job
    
    @classmethod
    def delete_job(cls, job_name: str, reason: str = "", workspace_id: Optional[str] = None) -> ExecutionJob:
        response = do_request(
            url="{}/v1/jobs/{}".format(
                Config.EXECUTION_BASE_URL_X_BACKEND,
                job_name
            ),
            params={
                "hard_delete": True,
                "reason": reason
            },
            http_method="delete",
            workspace_id=workspace_id
        )
        
        job_dict = response.json()["data"]
        job = ExecutionJob.parse_obj(job_dict)
        return job
    
    @classmethod
    def create_job(cls, job: ExecutionJob, reason: str = "", workspace_id: Optional[str] = None) -> ExecutionJob:        
        job_dict = job.dict()
        
        response = do_request(
            url="{}/v1/jobs/".format(
                Config.EXECUTION_BASE_URL_X_BACKEND
            ),
            params={
                "reason": reason
            },
            http_method="post",
            json_data=job_dict,
            workspace_id=workspace_id
        )
        
        returned_job_dict = response.json()["data"]
        returned_job = ExecutionJob.parse_obj(returned_job_dict)
        
        return returned_job
       


from typing import List, Any, Dict, Optional
from pydantic import BaseModel, validator
import numpy as np
import re
from xclientai.dtypes.shared import Status, Credentials, MachineType
from xclientai.utils.logging import configure_logger
from xclientai.config import Config
from urllib.parse import urlparse


logger = configure_logger(__name__)

    
class Signature(BaseModel):
    name: str
    shape: List[int]
    data: Optional[Any]
    datatype: str
    
    @validator("name")
    def name_is_valid(cls, v):
        assert isinstance(v, str), "Name property should be a string"        
        return v

    @validator("datatype")
    def datatype_is_valid(cls, v):
        valid_datatypes = [
            "int8",
            "int16",
            "int32",
            "int64",
            "fp16",
            "fp32",
            "fp64"
        ]

        assert v in valid_datatypes, f"datatype {v} is not valid. Use one of these: {valid_datatypes}"

        return v
    
    @validator("shape")
    def validate_shape(cls, v):
        return v

    @validator("data")
    def data_match_with_shape(cls, v, values, **kwargs):
        if v is not None:
            data_shape = list(np.array(v).shape)
            given_shape = values["shape"]

            assert len(data_shape) == len(given_shape), f"The data and the shape on the given input or output does not match: {values}"

            for given_dim, data_dim in zip(given_shape, data_shape):
                if given_dim != data_dim:
                    assert given_dim == -1, f"The data and the shape on the given input or output does not match: {values}"

        return v
    
    class Config:
        validate_all = True


class BenchmarkConfig(BaseModel):
    name: str
    dim: int
    resize_values: List[int]


class BenchmarkResults(BaseModel):
    baseline_model: Optional[List[Dict]]
    accelerated_model: Optional[List[Dict]]


class Accelerator(BaseModel):
    type: Optional[str]
    precision: Optional[str]
    
    @validator("type")
    def validate_accelerator_type(cls, v):
        valid_types = ["tensorrt"]
        
        assert v in valid_types, "{} is not a valid accelerator. Valid accelerators: {}".format(v, valid_types)
        
        return v
    
    @validator("precision")
    def validate_accelerator_precision(cls, v): 
        v = str(v) 
        valid_precisions = ["32", "16"]
        
        assert v in valid_precisions, "{} is not a valid accelerator precision. Valid accelerator precisions: {}".format(v, valid_precisions)
        
        return v

 
class BenchmarkInfo(BaseModel):
    benchmark_config: List[BenchmarkConfig]
    warmup: int = 3
    repeat: int = 3
    hardware: str = "gpu"
    accelerator: Optional[Accelerator]
    results: Optional[BenchmarkResults]
    
    @validator("hardware")
    def validate_hardware(cls, v):
        valid_hardware = ["cpu", "gpu"]
        assert v in valid_hardware, "{} is not a valid hardware. Valid hardwares: {}".format(v, valid_hardware)
        
        return v


class AccelerationInfo(BaseModel):
    input_model_path: str
    input_model_type: str
    output_model_path: str
    output_model_type: Optional[str]
    inputs: List[Signature]
    outputs: List[Signature]
    
    @validator("input_model_type")
    def validate_input_model_type(cls, v):
        valid_model_types = ["pt"]
        
        assert v in valid_model_types, "input_model_type should be one of the following values: {}".format(valid_model_types)
        
        return v


class Notification(BaseModel):
    headers: Optional[Dict]
    url: Optional[str]
    type: Optional[str]
    exclude_data_signature: bool = True
    
    @validator("url")
    def validate_url(cls, v):
        if v is not None:
            parsed_url_obj = urlparse(v)
            
            assert parsed_url_obj.scheme != '', "The URL should have the following format http://domain-name.com/path/. Don't forget the schema http or https"
        
        return v
    
    @validator("type")
    def validate_type(cls, v, values, **kwargs):
        valid_notification_methods = ["webhook"]
        assert v in valid_notification_methods, f"The valid notification methods are: {valid_notification_methods}"
        
        if v == "webhook":
            assert values.get("url") is not None, "If notification type is webhook, you should specify the field url with a string"
        
        return v


class NotificationResult(BaseModel):
    type: str
    message: str
    status_code: int

class AccelerationContainerSpecs(BaseModel):
    machine_type: MachineType = MachineType.GPU_T4_1
    spot: Optional[bool] = True

class AccelerationJob(BaseModel):
    job_name: str
    status: Status = Status.NOT_STARTED
    panel_url: Optional[str]
    error: Optional[str]
    notification_result: Optional[NotificationResult]
    container_specs: AccelerationContainerSpecs
    credentials: Credentials
    acceleration: AccelerationInfo
    benchmark: BenchmarkInfo = None
    notification: Notification = None
    started: Optional[int]
    ended: Optional[int]
    
    @validator("panel_url")
    def validate_panel_url(cls, v, values: dict, **kwargs):
        if values["status"] == Status.SUCCESSFUL:
            v = "{}/jobs/{}".format(
                Config.PANEL_DOMAIN,
                values["job_name"]
            )
        
        return v     
        
    @validator("job_name")
    def validate_job_name(cls, v):
        regex = "^[a-z0-9]([-a-z0-9]*[a-z0-9])?(\\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*$"
        
        assert len(v) < 30, "The job name cannot be longuer than 30 characters. Currently {}".format(len(v))
        
        assert bool(re.match(regex, v)), "The job_name must consist of lower case alphanumeric characters. Regex used for validation is '^[a-z0-9]([-a-z0-9]*[a-z0-9])?(\\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*$'"
        
        return v  
    
    @validator("benchmark")
    def validate_benchmark(cls, v: BenchmarkInfo, values: dict, **kwargs):
        n_dims = len(values["acceleration"].inputs[0].shape)
        
        for benchmark_config_item in v.benchmark_config:
            assert benchmark_config_item.dim < n_dims, "Benchmark dim cannot be greater than the number of dims in the input. Benchmark config: {}".format(benchmark_config_item)
            
        return v
    
    class Config:
        validate_all = True



from xclientai.__about__ import __version__

from xclientai.dtypes.shared import (
    Status,
    Credentials,
    MachineType
)

from xclientai.dtypes.optimization import (
    Signature,
    BenchmarkConfig,
    Accelerator,
    BenchmarkResults,
    BenchmarkInfo,
    AccelerationInfo,
    Notification,
    AccelerationJob
)

from xclientai.dtypes.executor import (
    ExecutionJob,
    ExecutionContainerSpecs,
    CodeSpecs
)

from xclientai.dtypes.deployments import (
    ModelSpecs,
    Batcher,
    Scaling,
    SCALE_METRIC,
    DeploymentSpecs,
    Deployment,
    DeploymentContainerSpecs
)

from xclientai.inference_utils import inference_utils

from xclientai.clients.acceleration_jobs import AccelerationJobsClient
from xclientai.clients.executor_jobs import ExecutionJobsClient
from xclientai.clients.deployments import DeploymentsClient


from typing import Dict, List, Optional, Union
from pydantic import BaseModel, validator, Field
import re
from xclientai.dtypes.shared import Status, Credentials, MachineType, Metadata
        
class CodeSpecs(BaseModel):
    credentials: Optional[Credentials]
    code_dir: Optional[str]
    
class ExecutionContainerSpecs(BaseModel):
    machine_type: MachineType = MachineType.GPU_T4_1
    spot: Optional[bool] = True
    image: str
    command: Optional[List[str]]
    args: Optional[List[str]]
    env: Optional[Dict[str, str]]
    secrets: Optional[Dict[str, str]] = Field(repr=False)
    
class ExecutionJob(BaseModel):
    job_name: str
    status: Status = Status.NOT_STARTED
    container_specs: List[ExecutionContainerSpecs]
    code_specs: Optional[CodeSpecs]
    metadata: Optional[Metadata]
    
    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        validate_all = True 
        
    @validator("job_name")
    def validate_job_name(cls, v):
        regex = "^[a-z0-9]([-a-z0-9]*[a-z0-9])?(\\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*$"
        
        assert len(v) <= 30, "The job name cannot be longuer than 30 characters. Currently {}".format(len(v))
        
        assert bool(re.match(regex, v)), "The job_name must consist of lower case alphanumeric characters. Regex used for validation is '^[a-z0-9]([-a-z0-9]*[a-z0-9])?(\\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*$'"
        
        return v



import os
import json
from xclientai.config import Config
from typing import Dict


def overwrite_config(xcloud_config: Dict):
    if not Config.XCLOUD_CONFIG_PATH.parent.is_dir():
        Config.XCLOUD_CONFIG_PATH.parent.mkdir(parents=True, exist_ok=True)
        
    with open(str(Config.XCLOUD_CONFIG_PATH), 'w', encoding='utf-8') as f:
        json.dump(xcloud_config, f, ensure_ascii=False, indent=4)

def read_config():
    if not Config.XCLOUD_CONFIG_PATH.is_file():
        return {}
        
    with open(Config.XCLOUD_CONFIG_PATH) as f:
        config = json.load(f)
        
    return config

def get_api_key():
    api_key = os.getenv("XCLOUD_API_KEY")
    
    if api_key is None:
        api_key = read_config().get("XCLOUD_API_KEY")
            
    return api_key

def get_default_workspace():
    default_workspace = os.getenv("XCLOUD_DEFAULT_WORKSPACE")
    
    if default_workspace is None:
        default_workspace = read_config().get("XCLOUD_DEFAULT_WORKSPACE")
            
    return default_workspace


import click

from xclientai.__about__ import __version__
from xclientai.cli.configure import configure_command


@click.group(
    context_settings={'help_option_names': ['-h', '--help']}, 
    invoke_without_command=True
)
@click.version_option(version=__version__, prog_name='xCloud')
@click.pass_context
def xcloud(ctx: click.Context):
    pass

xcloud.add_command(configure_command)



import requests
from requests.adapters import HTTPAdapter, Retry
import urllib3
from xclientai.utils.config_utils import get_api_key, get_default_workspace
from xclientai.config import Config
from xclientai.utils.logging import configure_logger
from typing import Optional

logger = configure_logger(__name__)

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


def do_request(
    url, 
    http_method, 
    json_data=None, 
    headers=None,
    params=None,
    xcloud_auth: bool = True,
    workspace_id: Optional[str] = None,
    throw_error: bool = True
):
    if xcloud_auth and not Config.DEBUG:
        api_key = get_api_key()
        assert api_key is not None, "We were not able to find your API key. To use the CLI or the Python client you need one. You can get an API Key in your profile. Then run the command `xcloud configure` or use the environment variable XCLOUD_API_KEY"
        
        headers = {} if headers is None else headers
        headers['x-api-key'] = api_key
        
    if workspace_id is None and not Config.DEBUG:
        workspace_id = get_default_workspace()
        
        assert workspace_id is not None, "The workspace_id cannot be None. Configure your workspace running the command `xcloud configure` or use the environment variable XCLOUD_DEFAULT_WORKSPACE"
        
        headers = {} if headers is None else headers
        headers['workspaceId'] = workspace_id
        
    session = requests.Session()
    retries = Retry(
        total=5,
        status_forcelist=[ 500, 502, 503, 504 ]
    )
    
    session.mount("http://", HTTPAdapter(max_retries=retries))
        
    params_fn = {
        "url": url, 
        "json": json_data,
        "timeout": 300,
        "headers": headers,
        "verify": False,
        "params": params
    }
    
    if Config.DEBUG:
        logger.debug(f"Request params: {params_fn}")
    
    if http_method == "get":
        response = session.get(**params_fn)
    elif http_method == "post":
        response = session.post(**params_fn)
    elif http_method == "patch":
        response = session.patch(**params_fn)
    elif http_method == "delete":
        response = session.delete(**params_fn)
    elif http_method == "put":
        response = session.put(**params_fn)
    
    error = False
    try:
        response.raise_for_status()
    except Exception as ex:
        error = True
        
    if error and throw_error:
        if 'application/json' in response.headers.get('Content-Type', ''):
            raise ValueError(response.json()["message"])
        else:
            raise ValueError(response.text)
    
    return response


from pathlib import Path
import tempfile
import os
from dotenv import load_dotenv, find_dotenv

# IMPORTANT: dont import logger here (circular import)

env_path = find_dotenv()
load_dotenv(env_path)


class Config:
    DEBUG = os.getenv("DEBUG", "false").lower() in ('true', '1', 't')
    ACCELERATION_BASE_URL_X_BACKEND = os.environ.get(
        "ACCELERATION_BASE_URL_X_BACKEND", 
        "https://xcloud-api.stochastic.ai/acceleron/backend"
    )
    EXECUTION_BASE_URL_X_BACKEND = os.environ.get(
        "EXECUTION_BASE_URL_X_BACKEND", 
        "https://xcloud-api.stochastic.ai/executor/backend"
    )
    DEPLOYMENTS_BASE_URL_X_BACKEND = os.environ.get(
        "DEPLOYMENTS_BASE_URL_X_BACKEND", 
        "https://xcloud-api.stochastic.ai/inference/backend"
    )
    PANEL_DOMAIN = "https://xcloud.stochastic.ai"
    AUTH_BASE_URL_X_BACKEND = os.environ.get(
        "AUTH_BASE_URL_X_BACKEND", 
        "https://xcloud-api.stochastic.ai/auth/backend"
    )
    XCLOUD_CONFIG_PATH: Path = Path.home() / ".xcloud" / "config"


import shutil
import zipfile
from pathlib import Path


def zip_dir(zip_file_name: str, dir_path: str):
    dir_path_object = Path(dir_path)
    assert dir_path_object.is_dir(), "dir_path should be a directory"
    shutil.make_archive(
        base_name=zip_file_name, 
        format='zip', 
        root_dir=str(dir_path_object.parent.resolve()), 
        base_dir=str(dir_path_object.name)
    )
    
def unzip_file(zip_file_path: str, dest_dir_path: str):
    zip_file_path_object = Path(zip_file_path).resolve()
    dest_dir_path_object = Path(dest_dir_path).resolve()
    
    assert zip_file_path_object.is_file(), "zip_file_path should be a file"
    assert dest_dir_path_object.is_dir(), "dest_dir_path should be a directory"
    
    with zipfile.ZipFile(str(zip_file_path_object), 'r') as zip_ref:
        zip_ref.extractall(str(dest_dir_path_object))



import setuptools
import os
from pathlib import Path
import re

from packaging.version import Version

version_string = os.environ.get("PKG_VERSION")
if version_string is None:
   raise Exception("PKG_VERSION not specified")

version = Version(version_string)
about_file = Path(__file__).parent.absolute() / "src/xclientai/__about__.py"

# read __about__.py file
with open(about_file, "r") as f:
   file_contents = f.read()

# replace __version__ = "0.0.0" string
version_line_code = f'__version__ = "{version}"'
new_contents = re.sub(
   r'^__version__ = ".+"$', 
   version_line_code, 
   file_contents
)

# write content to __about__.py file again
with open(about_file, "w") as f:
   f.write(new_contents)

print("NEW VERSION: {}".format(str(version)))

install_requires = [
   "setuptools",
   "pydantic",
   "numpy",
   "requests",
   "python-dotenv",
   "kserve",
   "click"
]


setuptools.setup(
   name='xclientai',
   version=str(version),
   author='Marcos Rivera Martínez, Sarthak Langde, Glenn Ko, Subhash G N, Toan Do, Roman Ageev',
   author_email='marcos.rm@stochastic.ai, sarthak.langde@stochastic.ai, glenn@stochastic.ai, subhash.gn@stochastic.ai',
   description='',
   long_description_content_type="text/markdown",
   url="",
   classifiers=[
      "Programming Language :: Python :: 3",
      "Operating System :: OS Independent"
   ],
   package_dir={"": "src"},
   packages=setuptools.find_packages(where="src"),
   python_requires=">=3.6",
   install_requires=install_requires,
   entry_points={
      "console_scripts": [
         "xcloud = xclientai.cli:xcloud"
      ]
   }  
)


from typing import List, Any, Dict, Optional
from pydantic import BaseModel, Field, validator
import re
from enum import Enum
from xclientai.dtypes.shared import Status, Credentials, MachineType, Metadata
    
   
class ModelSpecs(BaseModel):
    model_path: Optional[str]
    credentials: Optional[Credentials]

class Batcher(BaseModel):
    max_batch_size: Optional[int] = 1
    # Max latency in milliseconds
    max_lantecy: Optional[int] = 1
    # batch request timeout in milliseconds
    # 20 secs by default
    timeout: Optional[int] = 20000

class SCALE_METRIC(str, Enum):
    CONCURRENCY = "concurrency"
    RPS = "rps"

class Scaling(BaseModel):
    time_before_scaling_to_zero: Optional[int] = 1800000
    min_replicas: Optional[int] = 1
    max_replicas: Optional[int] = 1
    scale_metric: Optional[SCALE_METRIC] = SCALE_METRIC.CONCURRENCY
    target_scaling_value: Optional[int] = 1
    
    @validator("min_replicas")
    def validate_min_replicas(cls, v):
        assert v >= 0, "min_replicas value of the scaling config cannot be lower than 0"
            
        return v
    
    @validator("max_replicas")
    def validate_max_replicas(cls, v, values):
        assert v >= 0, "max_replicas value of the scaling config cannot be lower than 0"
        
        assert v >= values['min_replicas'], "max_replicas value of the scaling config cannot be lower than the min_replicas value"
            
        return v
    
    @validator("scale_metric")
    def validate_scale_metric(cls, v):
        valid_scale_metrics = [SCALE_METRIC.CONCURRENCY, SCALE_METRIC.RPS]
        
        assert v in valid_scale_metrics, "The allowed scale_metrics are SCALE_METRIC.CONCURRENCY, SCALE_METRIC.RPS"
        
        return v
    
    @validator("target_scaling_value")
    def validate_target_scaling_value(cls, v):
        assert v >= 0, "target_scaling_value value of the scaling config cannot be lower than 0"

        return v
    
    class Config:
        validate_all = True 

    
class DeploymentSpecs(BaseModel):
    batcher: Optional[Batcher] = Field(default_factory=Batcher)
    scaling: Optional[Scaling] = Field(default_factory=Scaling)
    
    
class Inference(BaseModel):
    base_endpoint: Optional[str]
    is_ready_endpoint: Optional[str]
    infer_endpoint: Optional[str]
    api_key: Optional[str]
    
    
class DeploymentContainerSpecs(BaseModel):
    machine_type: MachineType = MachineType.GPU_T4_1
    spot: Optional[bool] = True
    image: Optional[str]
    command: Optional[List[str]]
    args: Optional[List[str]]
    env: Optional[Dict[str, str]]
    secrets: Optional[Dict[str, str]] = Field(repr=False)
    

class Deployment(BaseModel):
    deployment_name: str
    status: Status = Status.NOT_STARTED
    model_specs: ModelSpecs
    deployment_specs: Optional[DeploymentSpecs] = Field(default_factory=DeploymentSpecs)
    container_specs: DeploymentContainerSpecs
    inference: Optional[Inference]
    metadata: Optional[Metadata]
    
    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        validate_all = True 
        
    @validator("deployment_name")
    def validate_deployment_name(cls, v):
        regex = "^[a-z0-9]([-a-z0-9]*[a-z0-9])?(\\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*$"
        
        assert len(v) <= 30, "The deployment name cannot be longuer than 30 characters. Currently {}".format(len(v))
        
        assert bool(re.match(regex, v)), "The deployment_name must consist of lower case alphanumeric characters. Regex used for validation is '^[a-z0-9]([-a-z0-9]*[a-z0-9])?(\\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*$'"
        
        return v   
